﻿using System;
using System.Collections.Generic;
using System.IO;
using System.Xml;
using Petzold.Epub.Common;
using Petzold.Epub.Css;

namespace Petzold.Epub.Html
{
    // At some point, HtmlFile has to have a property with number of breaks


    public class HtmlFile
    {
        Func<string, Stream> openFile;
        IFontMetricsProvider fontMetricsProvider;
        HtmlPageBuilder htmlPageBuilder;

        public HtmlFile(string source, string linear,
                        Func<string, Stream> openFile,
                        IFontMetricsProvider fontMetricsProvider)
        {
            bool isLinear;
            Boolean.TryParse(linear, out isLinear);

            this.Source = source;
            this.IsLinear = isLinear;
            this.openFile = openFile;
            this.fontMetricsProvider = fontMetricsProvider;
        }

        public bool IsLoaded { protected set; get; }

        public string Source { protected set; get; }

        public bool IsLinear { protected set; get; }

        public int TotalCharacterCount { set; get; }

        public HtmlElement Root { set; get; }

        public void Load()
        {
            if (this.IsLoaded)
                return;

            using (Stream stream = openFile(this.Source))
            {
                string html = null;

                using (StreamReader streamReader = new StreamReader(stream))
                {
                    html = streamReader.ReadToEnd();
                }

                html = html.Replace("&trade;", "&#x00AE;");
                html = html.Replace("&nbsp;", "&#x00A0;");

                using (StringReader stringReader = new StringReader(html))
                {
                    XmlReaderSettings readerSettings = new XmlReaderSettings
                    {
                        DtdProcessing = DtdProcessing.Ignore,
                        CheckCharacters = false
                    };

                    using (XmlReader xmlReader = XmlReader.Create(stringReader, readerSettings))
                    {
                        Load(xmlReader);
                    }
                }
            }

            htmlPageBuilder = new HtmlPageBuilder(this.Root, fontMetricsProvider);
            this.IsLoaded = true;
        }

        void Load(XmlReader xmlReader)
        {
            Stack<HtmlElement> elementStack = new Stack<HtmlElement>();
            HtmlElement parentElement = null;
            int accumulatedCharacterCount = 0;

            while (xmlReader.Read())
            {
                switch (xmlReader.NodeType)
                {
                    case XmlNodeType.Element:
                        HtmlElementType elementType = HtmlElementType._unsupported;
                        Enum.TryParse<HtmlElementType>(xmlReader.Name, out elementType);
                        HtmlElement thisElement = null;

                        if (elementType == HtmlElementType._unsupported)
                        {

                            System.Diagnostics.Debug.WriteLine("Unrecognized element type of " + xmlReader.Name);

                            thisElement = new HtmlElement(xmlReader.Name, parentElement, accumulatedCharacterCount);
                        }
                        else
                        {
                            thisElement = new HtmlElement(elementType, parentElement, accumulatedCharacterCount);
                        }

         //               System.Diagnostics.Debug.WriteLine("thisElement = {0} with parent = {1}", thisElement, parentElement);

                        // Add to parent's children collection
                        if (parentElement != null)
                        {
                            parentElement.AddChild(thisElement);
                        }
                        else // when thisElement is 'html'
                        {
                            this.Root = thisElement;
                        }

                        // Most famous empty element is 'br'
                        if (!xmlReader.IsEmptyElement)
                        {
             //               System.Diagnostics.Debug.WriteLine("    thisElement {0} pushed on stack", thisElement);
                            elementStack.Push(thisElement);
                            parentElement = thisElement;
                        }

                        while (xmlReader.MoveToNextAttribute())
                        {
                            thisElement.AddAttribute(xmlReader.Name, xmlReader.Value);
                        }
                        break;

                    case XmlNodeType.Text:
                        // TODO: Determine if removing control codes (eg, from Gutenberg files) is the correct solution here
                        string text = xmlReader.Value;
                        text = text.Replace("\r", "");
                        text = text.Replace("\n", "");

                        // Create HtmlElement object and add to parent's collection
                        HtmlElement textElement = new HtmlElement(parentElement, text, accumulatedCharacterCount);
                        parentElement.AddChild(textElement);

                        // Bump up the accumulatedCharacterCount
                        accumulatedCharacterCount += text.Length;
                        break;

                    case XmlNodeType.EndElement:
                        HtmlElementType endElementType = HtmlElementType._unsupported;
                        Enum.TryParse<HtmlElementType>(xmlReader.Name, out endElementType);

                        if (elementStack.Count == 0)
                            throw new Exception("ElementStack prematurely ran out!");

                        HtmlElement poppedElement = elementStack.Pop();

                //        System.Diagnostics.Debug.WriteLine("EndElementType = {0} parentElement = {1} poppedElement = {2}", endElementType, parentElement, poppedElement);

                        if (parentElement != poppedElement) //  || thisElement.ElementType != endElementType)
                        {
                            throw new ArgumentException("HtmlParser can't match end-element tag with start-element tag");
                        }

                        parentElement = (elementStack.Count > 0) ? elementStack.Peek() : null;
                        break;

                    case XmlNodeType.XmlDeclaration:
          //              System.Diagnostics.Debug.WriteLine("Ignoring xmlDeclaration of name = {0} value = {1}.", xmlReader.Name, xmlReader.Value);
                        break;

                    case XmlNodeType.Whitespace:
         //               System.Diagnostics.Debug.WriteLine("Ignoring whitespace of name = {0} value = {1}.", xmlReader.Name, xmlReader.Value);
                        break;

                    case XmlNodeType.Comment:
            //            System.Diagnostics.Debug.WriteLine("Ignoring comment of name = {0} value = {1}.", xmlReader.Name, xmlReader.Value);
                        break;

                    default:
                        throw new ArgumentException("HtmlFile not processing NodeType " + xmlReader.NodeType);
                }
            }
            this.TotalCharacterCount = accumulatedCharacterCount;
        }


        public PageLayout GetPageLayout(PageStart pageStart, out PageStart nextPageStart, PageParams pageParams)
        {
            return htmlPageBuilder.GetPageLayout(pageStart, out nextPageStart, pageParams);
        }



        internal static IList<int> GetElementIndices(HtmlElement htmlElement)
        {
            List<int> indices = new List<int>();

            while (htmlElement.Parent != null)
            {
                HtmlElement parent = htmlElement.Parent;
                indices.Insert(0, parent.Children.IndexOf(htmlElement));
                htmlElement = parent;
            }

            return indices;
        }

        internal static HtmlElement FindElement(HtmlElement rootElement, IList<int> indices)
        {
            HtmlElement htmlElement = rootElement;

            if (indices != null)
            {
                foreach (int index in indices)
                {
                    if (htmlElement.Children == null)
                        return null;

                    if (htmlElement.Children.Count <= index)
                        return null;

                    htmlElement = htmlElement.Children[index];
                }
            }
            return htmlElement;
        }


 


    }
}
