﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Threading.Tasks;
using Petzold.Epub.Common;
using Petzold.Epub.Css;

namespace Petzold.Epub.Html
{
    internal class HtmlPageBuilder
    {
        // Information from constructor
        HtmlElement rootElement;
        IFontMetricsProvider fontMetricsProvider;

        PageLayout pageLayout;
        HtmlElement currentElement;
        int textIndex;
        double x, y;

        int lineBeginContentCount = 0;
        HtmlElement lineBeginElement = null;
        int lineBeginTextIndex = 0;


        // TODO: Maybe create new HtmlPageBuilder object for each page -- would certainly simplify initialization


        public HtmlPageBuilder(HtmlElement rootElement, IFontMetricsProvider fontMetricsProvider)
        {
            this.rootElement = rootElement;
            this.fontMetricsProvider = fontMetricsProvider;
        }



        public PageLayout GetPageLayout(PageStart pageStart, out PageStart nextPageStart, PageParams pageParams)
        {
            // Initialize fields
            currentElement = HtmlFile.FindElement(rootElement, pageStart.ElementIndices);
            textIndex = pageStart.TextIndex;

            // Create PageLayout with character offset
            pageLayout = new PageLayout(pageParams.PageSize, pageStart.CharacterOffset); // or (currentElement.AccumulatedCharacterCount + textIndex);


            // Initialize coordinates
            x = 0;
            y = 0;

            // Initialize rewind information
            lineBeginContentCount = 0;
            lineBeginElement = null;
            lineBeginTextIndex = 0;

            // For convenience
            double pageWidth = pageParams.PageSize.Width;
            double pageHeight = pageParams.PageSize.Height;


            while (y < pageHeight)
            {
                bool needNextElement = false;

                // TODO: Don't go too long ignoring img!
                if (currentElement.ElementType == HtmlElementType.img)
                {
                    string imgSrc = null;
                    string imgWidth = null;
                    string imgHeight = null;

                    if (currentElement.Attributes.ContainsKey("src"))
                        imgSrc = currentElement.Attributes["src"];

                    if (currentElement.Attributes.ContainsKey("width"))
                        imgWidth = currentElement.Attributes["width"];

                    if (currentElement.Attributes.ContainsKey("height"))
                        imgHeight = currentElement.Attributes["height"];

                    Word word = new Word
                    {
                        Text = String.Format("img src = {0} width = {1} height = {2}", imgSrc, imgWidth, imgHeight),
                        X = x,
                        Y = y,
                        FontSize = 24,
                        Color = new RgbColor(255, 0, 0),
                    };
                    y += fontMetricsProvider.GetIntegerSize(word).Height;
                    pageLayout.Content.Add(word);
                    needNextElement = true;
                }

                // TODO: This is temporary to provide a line break for common block elements
                //      (Not doing anything for img, table, or lists
                else if (currentElement.ElementType == HtmlElementType.br ||
                    currentElement.ElementType == HtmlElementType.div ||
                    currentElement.ElementType == HtmlElementType.p ||
                    currentElement.ElementType == HtmlElementType.address ||
                    currentElement.ElementType == HtmlElementType.blockquote ||
                    currentElement.ElementType == HtmlElementType.pre ||
                    currentElement.ElementType == HtmlElementType.h1 ||
                    currentElement.ElementType == HtmlElementType.h2 ||
                    currentElement.ElementType == HtmlElementType.h3 ||
                    currentElement.ElementType == HtmlElementType.h4 ||
                    currentElement.ElementType == HtmlElementType.h5 ||
                    currentElement.ElementType == HtmlElementType.h6)
                {
                    DropToNextLine(fontMetricsProvider.GetIntegerSize(new Word { Text = " ", FontSize = pageParams.DefaultFontSize }).Height);
                    needNextElement = true;
                }

                else if (currentElement.ElementType == HtmlElementType._text)
                {
                    int saveTextIndex = textIndex;
                    string wordText = GetNextWord(currentElement.Text, ref textIndex);

                    if (wordText != null)
                    {
                        Word word = new Word
                        {
                            Text = wordText,
                            FontSize = pageParams.DefaultFontSize
                        };

                        TextSize textSize = fontMetricsProvider.GetIntegerSize(word);

                        if (x + textSize.Width > pageWidth)
                        {
                            DropToNextLine(textSize.Height);

                            if (y > pageHeight)
                            {
                                textIndex = saveTextIndex;
                                break;
                            }
                        }

                        // Add Word object to content but watch out for
                        //  page overruns (but not if this is the only line)!
                        if (y + textSize.Height > pageHeight && lineBeginContentCount > 0)
                        {
                            // Rewind to beginning of line
                            while (pageLayout.Content.Count > lineBeginContentCount)
                                pageLayout.Content.RemoveAt(pageLayout.Content.Count - 1);

                            currentElement = lineBeginElement;
                            textIndex = lineBeginTextIndex;
                            break;                              // drops to bottom of method
                        }

                        word.X = x;
                        word.Y = y;
                        pageLayout.Content.Add(word);
                        x += textSize.Width + textSize.SpaceWidth;

                        if (x >= pageWidth)
                        {
                            // This will eventually be the largest text height in the line
                            DropToNextLine(textSize.Height);

                            if (y > pageHeight)
                                break;
                        }
                    }
                    else
                    {
                        needNextElement = true;
                    }
                }
                else
                {
                    needNextElement = true;
                }

                if (needNextElement)
                {
                    textIndex = 0;

                    if (currentElement.Children != null && currentElement.Children.Count > 0)
                    {
                        currentElement = currentElement.Children[0];
                    }
                    else
                    {
                        HtmlElement sibling = currentElement.NextSibling;

                        if (sibling != null)
                        {
                            currentElement = sibling;
                        }
                        else
                        {
                            while (true)
                            {
                                HtmlElement parent = currentElement.Parent;

                                if (parent.ElementType == HtmlElementType.body)
                                {
                                    pageStart.IsLastPage = true;
                                    pageLayout.IsLastPage = true;
                                    nextPageStart = null;
                                    return pageLayout;
                                }

                                sibling = parent.NextSibling;

                                if (sibling != null)
                                {
                                    currentElement = sibling;
                                    break;
                                }
                                else
                                {
                                    currentElement = parent;
                                }
                            }
                        }
                    }
                }
            }

            // Create new PageStart object for next page
            IList<int> indices = HtmlFile.GetElementIndices(currentElement);
            int characterOffset = currentElement.AccumulatedCharacterCount + textIndex;
            nextPageStart = new PageStart(indices, textIndex, characterOffset);
            return pageLayout;
        }

        string GetNextWord(string text, ref int textIndex)
        {
            if (String.IsNullOrEmpty(text))
                return null;

            while (textIndex < text.Length && text[textIndex] == ' ')
            {
                textIndex += 1;
            }

            if (textIndex == text.Length)
                return null;

            string word = null;
            int spaceIndex = text.IndexOf(' ', textIndex);

            if (spaceIndex == -1)
            {
                word = text.Substring(textIndex);
            }
            else
            {
                word = text.Substring(textIndex, spaceIndex - textIndex);
            }

            textIndex += word.Length;
            return word;
        }

        void DropToNextLine(double lineHeight)
        {
            y += lineHeight;
            x = 0;
            lineBeginContentCount = pageLayout.Content.Count;
            lineBeginElement = currentElement;
            lineBeginTextIndex = textIndex;
        }

    }
}
