﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Threading.Tasks;
using Petzold.Epub.Common;
using Windows.UI.Xaml;
using Windows.UI.Xaml.Input;

namespace Petzold.Epub.Viewer.Windows8
{
    public partial class BookViewer
    {
        bool canGoNextPage, canGoPrevPage;
        double dragFactor;

        // First event: set Booleans for going to next or previous page
        protected override void OnManipulationStarting(ManipulationStartingEventArgs args)
        {
            canGoNextPage = pageHosts[(pageHostBaseIndex + 1) % 3].Child != null;
            canGoPrevPage = pageHosts[(pageHostBaseIndex + 2) % 3].Child != null;

            args.Handled = true;
            base.OnManipulationStarting(args);
        }

        // If Tapped, this is the end of the line
        protected override void OnTapped(TappedEventArgs args)
        {
            if (canGoNextPage)
            {
                AnimateToNextPage();
            }

            args.Handled = true;
            base.OnTapped(args);
        }

        // If a drag has started, initialize the drag factor
        protected override void OnManipulationStarted(ManipulationStartedEventArgs args)
        {
            dragFactor = 0;
            args.Handled = true;
            base.OnManipulationStarted(args);
        }

        protected override void OnManipulationDelta(ManipulationDeltaEventArgs args)
        {
            double newDragFactor = dragFactor - args.DeltaManipulation.Translation.X / pageHosts[pageHostBaseIndex].ActualWidth;

            // Try getting the last page and reset canGoPrevPage
            if (newDragFactor < 0 && !canGoPrevPage && !this.CurrentLocation.IsFirstPage)
            {
                PageRequest pageRequest = new PageRequest
                {
                    Type = PageRequestType.GetLastPage,
                    Location = this.CurrentLocation.PreviousPage()
                };

                PageLayout pageLayout = PageProvider.GetPage(pageRequest, pageParams);
                pageHosts[(pageHostBaseIndex + 2) % 3].Child = BuildPage(pageLayout);
                canGoPrevPage = pageHosts[(pageHostBaseIndex + 2) % 3].Child != null;
            }

            double maximum = canGoNextPage ? 1 : 0;
            double minimum = canGoPrevPage ? -1 : 0;
            dragFactor = Math.Max(minimum, Math.Min(maximum, newDragFactor));
            this.PageTransition.FractionalBaseIndex = pageHostBaseIndex + dragFactor;

            args.Handled = true;
            base.OnManipulationDelta(args);
        }

        protected override void OnManipulationInertiaStarting(ManipulationInertiaStartingEventArgs args)
        {
            base.OnManipulationInertiaStarting(args);
        }

        protected override void OnManipulationCompleted(ManipulationCompletedEventArgs e)
        {
            if (dragFactor > 0.5)
                AnimateToNextPage();

            else if (dragFactor < -0.5)
                AnimateToPrevPage();

            else
                AnimateToSamePage();

            base.OnManipulationCompleted(e);
        }
        

        void AnimateToNextPage()
        {
            DoubleAnimator doubleAnimator = new DoubleAnimator
                {
                    Target = this.PageTransition,
                    Property = PageTransition.FractionalBaseIndexProperty,
                    ToValue = pageHostBaseIndex + 1,
                    Duration = CalculateFractionalDuration(pageHostBaseIndex + 1),

                    Completed = (sender, args) =>
                        {
                            pageHostBaseIndex = (pageHostBaseIndex + 1) % 3;
                            this.PageTransition.FractionalBaseIndex = pageHostBaseIndex;

                            // Transfer the indices from current to previous, and from next to current
                            prevPageLocation = this.CurrentLocation;
                            this.CurrentLocation = nextPageLocation;

                            // Setup the next page
                            SetupNextPage();
                        }
                };
            doubleAnimator.Begin();
            return;
        }

        void AnimateToPrevPage()
        {
            DoubleAnimator doubleAnimator = new DoubleAnimator
            {
                Target = this.PageTransition,
                Property = PageTransition.FractionalBaseIndexProperty,
                ToValue = pageHostBaseIndex - 1,
                Duration = CalculateFractionalDuration(pageHostBaseIndex - 1),

                Completed = (sender, args) =>
                {
                    pageHostBaseIndex = (pageHostBaseIndex + 2) % 3;
                    this.PageTransition.FractionalBaseIndex = pageHostBaseIndex;

                    // Transfer the indices from current to next, and from previous to current
                    nextPageLocation = this.CurrentLocation;
                    this.CurrentLocation = prevPageLocation;

                    // Setup the previous page
                    SetupPrevPage();
                }
            };
            doubleAnimator.Begin();
            return;
        }

        void AnimateToSamePage()
        {
            DoubleAnimator doubleAnimator = new DoubleAnimator
            {
                Target = this.PageTransition,
                Property = PageTransition.FractionalBaseIndexProperty,
                ToValue = pageHostBaseIndex,
                Duration = CalculateFractionalDuration(pageHostBaseIndex),
            };
            doubleAnimator.Begin();
            return;
        }

        Duration CalculateFractionalDuration(double fractionalBaseIndexDestination)
        {
            double fraction = Math.Abs(fractionalBaseIndexDestination - this.PageTransition.FractionalBaseIndex);
            double msec = fraction * this.PageTransition.AnimationDuration;
            TimeSpan timeSpan = TimeSpan.FromMilliseconds(msec);
            return new Duration { Type = DurationType.TimeSpan, TimeSpan = timeSpan };
        }
    }
}
