﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Threading.Tasks;
using Windows.Foundation;
using Windows.UI.Xaml;
using Windows.UI.Xaml.Controls;
using Windows.UI.Xaml.Data;
using Windows.UI.Xaml.Media;
using Petzold.Epub;
using Petzold.Epub.Common;
using Windows.UI.Xaml.Input;
using Windows.UI.Xaml.Media.Animation;

namespace Petzold.Epub.Viewer.Windows8
{
    public partial class BookViewer
    {
        Border[] pageHosts = new Border[3];
        int pageHostBaseIndex = 0;
        BookLocation nextPageLocation, prevPageLocation;
        PageParams pageParams = new PageParams();

        public static readonly DependencyProperty PageProviderProperty =
            DependencyProperty.Register("PageProvider",
                "Object",                                   // TODO: typeof(IPageProvider)
                typeof(BookViewer).FullName,
                new PropertyMetadata(null, OnPageProviderChanged));

        public static readonly DependencyProperty PageTransitionProperty =
            DependencyProperty.Register("PageTransition",
                "Object",                                   // TODO: typeof(PageTransition),
                typeof(BookViewer).FullName,
                new PropertyMetadata(null, OnPageTransitionChanged));

        public static readonly DependencyProperty CurrentLocationProperty =
            DependencyProperty.Register("CurrentLocation",
                "Object",                                   // TODO: typeof(BookLocation),
                typeof(BookViewer).FullName,
                new PropertyMetadata(0, OnCurrentLocationChanged));

        // Why can't this be an event?
        public Windows.UI.Xaml.EventHandler CurrentLocationChanged { set; get; }


        public BookViewer()
        {
            InitializeComponent();

            pageHosts[0] = pageHost0;
            pageHosts[1] = pageHost1;
            pageHosts[2] = pageHost2;

            // All touch events are handled in BookViwer.Touch.cs
            this.ManipulationMode = ManipulationModes.TranslateX | 
                                    ManipulationModes.TranslateY | 
                                    ManipulationModes.TranslateInertia;
        }

        public IPageProvider PageProvider
        {
            set { SetValue(PageProviderProperty, value); }
            get { return (IPageProvider)GetValue(PageProviderProperty); }
        }

        public PageTransition PageTransition
        {
            set { SetValue(PageTransitionProperty, value); }
            get { return (PageTransition)GetValue(PageTransitionProperty); }
        }

        public BookLocation CurrentLocation
        {
            protected set { SetValue(CurrentLocationProperty, value); }
            get { return (BookLocation)GetValue(CurrentLocationProperty); }
        }

        private void OnManipulationGridSizeChanged(object sender, SizeChangedEventArgs args)
        {
            if (pageHosts[pageHostBaseIndex].Child == null)
                return;

            // Get rid of all existing pages
            for (int i = 0; i < 3; i++)
                pageHosts[i].Child = null;

            pageHostBaseIndex = 0;

            JumpToPage(this.CurrentLocation.FileIndex, this.CurrentLocation.BreakIndex, this.CurrentLocation.CharacterOffset);
        }

        // Change arguments to BookLocation?
        public bool JumpToPage(int fileIndex, int breakIndex, int charOffset)
        {
            if (this.PageProvider == null)
                return false;

            pageParams.PageSize = new Dimension(pageHost0.ActualWidth, pageHost0.ActualHeight);
            pageParams.DefaultFontSize = 18;

            PageRequest pageRequest = new PageRequest
            {
                Type = PageRequestType.CharacterOffset,
                Location = new BookLocation(fileIndex, breakIndex, 0, charOffset)
            };

            PageLayout pageLayout = this.PageProvider.GetPage(pageRequest, pageParams);

            if (pageLayout != null)
            {
                CurrentLocation = new BookLocation(fileIndex, breakIndex, pageRequest.Location.PageIndex, charOffset);

                FrameworkElement pageElement = BuildPage(pageLayout);
                pageHosts[pageHostBaseIndex].Child = pageElement;
                SetupNextPage();
                SetupPrevPage();

                // Set the FractionalBaseIndex for the PageTransition
                this.PageTransition.FractionalBaseIndex = pageHostBaseIndex;
                return true;
            }            
            return false;
        }

        void SetupNextPage()
        {
            // Try getting the next page
            PageRequest pageRequest = new PageRequest
            {
                Type = PageRequestType.PageIndex,
                Location = this.CurrentLocation.NextPage()
            };

            PageLayout pageLayout = this.PageProvider.GetPage(pageRequest, pageParams);

            // TODO: Need to check next break in file

            // Gone beyond end of file so try next file
            if (pageLayout == null)
            {
                pageRequest.Type = PageRequestType.PageIndex;
                pageRequest.Location = this.CurrentLocation.NextFile();

                pageLayout = this.PageProvider.GetPage(pageRequest, pageParams);
            }

            if (pageLayout != null)
            {
                nextPageLocation = pageRequest.Location;
            }

            // Doesn't matter if pageLayout is null; need to set null Child in that case
            pageHosts[(pageHostBaseIndex + 1) % 3].Child = BuildPage(pageLayout);
        }

        void SetupPrevPage()
        {
            // Get out quickly for the only case where there's not a previous page
            if (this.CurrentLocation.IsFirstPage)
            {
                pageHosts[(pageHostBaseIndex + 2) % 3].Child = null;
                prevPageLocation = this.CurrentLocation.PreviousPage(); // sets all fields to -1
                return;
            }

            PageLayout pageLayout = null;
            PageRequest pageRequest = new PageRequest
            {
                Type = PageRequestType.PageIndex,
                Location = this.CurrentLocation.PreviousPage()
            };

            // Try getting previous page
            if (pageRequest.Location.PageIndex >= 0)
            {
                pageLayout = this.PageProvider.GetPage(pageRequest, pageParams);
            }
            // or last page of previous file
            else
            {
                pageRequest.Type = PageRequestType.QueryLastPage;
                pageLayout = this.PageProvider.GetPage(pageRequest, pageParams);
            }

            if (pageLayout != null)
            {
                prevPageLocation = pageRequest.Location;
            }
            pageHosts[(pageHostBaseIndex + 2) % 3].Child = BuildPage(pageLayout);
        }

        // Platform-independent PageLayout --> platform-dependent element
        FrameworkElement BuildPage(PageLayout pageLayout)
        {
            if (pageLayout == null)
                return null;

            Canvas pageCanvas = new Canvas
                {
                    Background = new SolidColorBrush(Colors.White),
                    Width = pageLayout.PageSize.Width,
                    Height = pageLayout.PageSize.Height
                };

            foreach (PageItem item in pageLayout.Content)
            {
                if (item is Word)
                {
                    Word word = item as Word;

                    TextBlock txtblk = CreateTextBlock(word);
                    Canvas.SetLeft(txtblk, word.X);
                    Canvas.SetTop(txtblk, word.Y);
                    pageCanvas.Children.Add(txtblk);
                }
            }
            return pageCanvas;
        }

        // Almost the same as the similar method in FontMetricsProvider
        TextBlock CreateTextBlock(Word word)
        {
            TextBlock txtblk = new TextBlock
            {
                FontStyle = word.IsItalic ? FontStyles.Italic : FontStyles.Normal,
                FontWeight = word.IsBold ? FontWeights.Bold : FontWeights.Normal,
                Text = word.Text,
                Foreground = new SolidColorBrush(ColorHelper.FromArgb(255, word.Color.R, word.Color.G, word.Color.B))
            };

            if (word.FontSize > 0)
                txtblk.FontSize = word.FontSize;

            if (!String.IsNullOrEmpty(word.FontFamily))
                txtblk.FontFamily = new FontFamily { Source = word.FontFamily };

            return txtblk;
        }

        #region Changed handlers for dependency properties

        static void OnPageProviderChanged(DependencyObject obj, DependencyPropertyChangedEventArgs args)
        {
        }
        
        static void OnPageTransitionChanged(DependencyObject obj, DependencyPropertyChangedEventArgs args)
        {
            (obj as BookViewer).OnPageTransitionChanged((PageTransition)args.OldValue, (PageTransition)args.NewValue);
        }

        void OnPageTransitionChanged(PageTransition oldPageTransition, PageTransition newPageTransition)
        {
            double fractionalBaseIndex = -1;

            if (oldPageTransition != null)
            {
                fractionalBaseIndex = oldPageTransition.FractionalBaseIndex;
                oldPageTransition.Detach();
            }
            if (newPageTransition != null)
            {
                newPageTransition.Attach(LayoutRoot, pageContainer0, pageContainer1, pageContainer2);
                newPageTransition.FractionalBaseIndex = fractionalBaseIndex;
            }
        }

        static void OnCurrentLocationChanged(DependencyObject obj, DependencyPropertyChangedEventArgs args)
        {
            BookViewer bookViewer = obj as BookViewer;

            if (bookViewer.CurrentLocationChanged != null)
                bookViewer.CurrentLocationChanged(bookViewer, EventArgs.Empty);
        }

        #endregion
    }
}
