﻿using System;
using System.Collections.Generic;
using System.IO;
using System.Linq;
using System.Text;
using System.Threading.Tasks;
using epub = Petzold.Epub.Package.EpubContainerFile;
using opf = Petzold.Epub.Package.OpfPackageDocument;
using nav = Petzold.Epub.Package.NcxNavigationDocument;
using Petzold.Epub.Html;

namespace Petzold.Epub
{
    public partial class EpubBook
    {
        public static IList<EpubBook> GetAllBooksInPackage(IPackageAccessor packageAccessor)
        {
            List<EpubBook> books = new List<EpubBook>();

            using (Stream mimetypeStream = packageAccessor.GetFile("mimetype"))
            {
                if (mimetypeStream == null)
                    throw new EpubException("The EPUB file does not include a required mimetype entry");

                using (StreamReader reader = new StreamReader(mimetypeStream))
                {
                    string strMimeType = reader.ReadToEnd();

                    if (strMimeType != "application/epub+zip")
                        throw new EpubException("The EPUB file does not include a valid mime type");
                }
            }

            using (Stream containerStream = packageAccessor.GetFile("META-INF/container.xml"))
            {
                if (containerStream == null)
                    throw new EpubException("The EPUB file does not include a required container entry");

                epub.container container = null;

                try
                {
                    container = epub.container.Deserialize(containerStream); // reader);
                }
                catch (Exception exc)
                {
                    throw new EpubException("The EPUB file contains a corrupted container entry", exc);
                }

                foreach (var rootfile in container.rootfiles.rootFileArray)
                {
                    if (rootfile.mediatype == "application/oebps-package+xml")
                    {
                        EpubBook book = new EpubBook(rootfile.fullpath, packageAccessor, null);
                        books.Add(book);
                    }
                }
            }
            return books;
        }

        void ParseOpfFile()
        {
            using (Stream packageStream = packageAccessor.GetFile(this.PackageFilePath))
            {
                if (packageStream == null)
                    throw new EpubException("The EPUB file is missing a package file for the book");

                opf.package package = null;

                try
                {
                    package = opf.package.Deserialize(packageStream);
                }
                catch (Exception exc)
                {
                    throw new EpubException("The package file for the book is corrupted", exc);
                }

                // Set bookInformation field for BookInformation property
                bookInformation = new BookInformation(package.metadata);

                // Enumerate the files in the manifest to get the NCX file
                foreach (opf.item item in package.manifest.items)
                {
                    // NOTE: The "toc" attribute of "spine" is required, but some EPUBs don't have it.
                    if (item.mediaType == "application/x-dtbncx+xml" &&
                            (String.IsNullOrEmpty(package.spine.toc) || (item.id == package.spine.toc)))
                    {
                        navigationFilePath = item.href;
                        break;
                    }
                }

                // Accumulate the XHTML files in the manifest
                Dictionary<string, string> manifestDictionary = new Dictionary<string, string>();

                foreach (opf.item item in package.manifest.items)
                {
                    // Save the XHTML files that comprise the book in the dictionary
                    if (item.mediaType == "application/xhtml+xml")
                    {
                        manifestDictionary.Add(item.id, item.href);
                    }
                }

                // Now put together the list of XHTML files that comprise the book
                htmlFiles = new List<HtmlFile>();

                foreach (opf.itemref itemref in package.spine.itemrefs)
                {
                    if (!manifestDictionary.ContainsKey(itemref.idref))
                        throw new EpubException(String.Format("idref of {0} in the spine is not included in the manifest", itemref.idref));

                    htmlFiles.Add(new HtmlFile(manifestDictionary[itemref.idref], itemref.linear, OpenFile, fontMetricsProvider));
                }

                if (String.IsNullOrEmpty(navigationFilePath))
                    throw new EpubException("The EPUB file is missing a navigation file for the book");
            }
        }

        void ParseNcxFile()
        {
            using (Stream navigationStream = packageAccessor.GetFile(this.navigationFilePath))
            {
                if (navigationStream == null)
                    throw new EpubException("The EPUB file is missing a navigation file for the book");

                nav.ncx ncx = null;

                try
                {
                    ncx = nav.ncx.Deserialize(navigationStream);
                }
                catch (Exception exc)
                {
                    throw new EpubException("The navigation file for the book is corrupted", exc);
                }

                this.tableOfContents = new TableOfContents(ncx.navMap);
            }
        }
    }
}
