﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Threading.Tasks;
using Petzold.Epub.Common;
using Petzold.Epub.Html;

namespace Petzold.Epub
{
    public class PageProvider: IPageProvider
    {
        IList<HtmlFile> htmlFiles;

        // For pagestarts, we need an array based on number of files and breaks referencing a dictionary with a key based on PageParams. 
        //  An item in the dictionary is a List<PageStart> for each consecutive page in the file that's been paginated.
        // For now, restrict it to files 
        Dictionary<PageParams, List<PageStart>>[] pageStarts; 

        public PageProvider(IList<HtmlFile> htmlFiles)
        {
            this.htmlFiles = htmlFiles;

            // Initialize pageStarts array
            pageStarts = new Dictionary<PageParams, List<PageStart>>[this.htmlFiles.Count];
        }

        public PageLayout GetPage(PageRequest pageRequest, PageParams pageParams)
        {
            BookLocation location = pageRequest.Location;

            // Check if the FileIndex exceeds the number of files
            if (location.FileIndex >= htmlFiles.Count)
            {
                System.Diagnostics.Debug.WriteLine("PageProvider.GetPage has a too-high FileIndex");
                return null;
            }
            // Get the HtmlFile
            HtmlFile htmlFile = htmlFiles[pageRequest.Location.FileIndex];

            // TODO: Name this method something other than Load.  Parse, perhaps?
            // Should we put it in a try/catch ?
            htmlFile.Load();
            // Loading the HtmlFile should also provide a number of breaks, hence the PageRequest can use this information

            // Check if there's already a Dictionary entry in pageStarts
            if (pageStarts[location.FileIndex] == null)
                pageStarts[location.FileIndex] = new Dictionary<PageParams, List<PageStart>>();


            // Initialize the dictionary with a PageStart item indicating the first element in 'body'
            if (!pageStarts[location.FileIndex].ContainsKey(pageParams))
            {
                pageStarts[location.FileIndex].Add(pageParams, new List<PageStart>(new PageStart[] { new PageStart(new int[] { 1, 0 }, 0, 0) }));
            }
            // Get the List of PageStarts
            List<PageStart> pageStartList = pageStarts[location.FileIndex][pageParams];

            PageStart pageStart = null;
            PageLayout pageLayout = null;
            PageStart nextPageStart;        // Created by call to HtmlFile.GetPageLayout (which goes to HtmlPageBuilder.GetPageLayout)

            switch (pageRequest.Type)
            {

                // TODO: Logic here needs some careful analysis!
                case PageRequestType.CharacterOffset:
                    int pageIndex = 0;

                    while (!pageStartList[pageIndex].IsLastPage)
                    {
                        if (pageIndex == pageStartList.Count - 1)
                        {
                            pageStart = pageStartList[pageIndex];

                            pageLayout = htmlFile.GetPageLayout(pageStart, out nextPageStart, pageParams);

                            // These two if's are actually equivalent
                            if (nextPageStart != null)
                                pageStartList.Add(nextPageStart);

                            if (pageStartList[pageIndex].IsLastPage)
                                break;
                        }
                        if (location.CharacterOffset < pageStartList[pageIndex + 1].CharacterOffset)
                        {
                            pageStart = pageStartList[pageIndex];
                            pageLayout = htmlFile.GetPageLayout(pageStart, out nextPageStart, pageParams);
                            break;
                        }
                        pageIndex += 1;
                    }

                    if (pageStartList[pageIndex].IsLastPage)
                    {
                        pageStart = pageStartList[pageIndex];
                        pageLayout = htmlFile.GetPageLayout(pageStart, out nextPageStart, pageParams);
                    }

                    pageRequest.Location = new BookLocation(location.FileIndex, location.BreakIndex, pageIndex, pageStart.CharacterOffset);
                    break;

                case PageRequestType.PageIndex:
                    if (location.PageIndex >= pageStartList.Count)
                    {
                        System.Diagnostics.Debug.WriteLine("PageProvider.GetPage has a too-high PageIndex");
                        return null;
                    }
                    pageStart = pageStartList[location.PageIndex];
                    pageLayout = htmlFile.GetPageLayout(pageStart, out nextPageStart, pageParams);

                    // Add to the end of the list only if we're currently getting the last known page
                    if (nextPageStart != null && pageStartList.Count == location.PageIndex + 1)
                        pageStartList.Add(nextPageStart);

                    pageRequest.Location = new BookLocation(location.FileIndex, location.BreakIndex, location.PageIndex, pageStart.CharacterOffset);
                    break;

                case PageRequestType.QueryLastPage:
                    if (pageStartList[pageStartList.Count - 1].IsLastPage)
                        goto case PageRequestType.GetLastPage;

                    break;

                case PageRequestType.GetLastPage:
                    do
                    {
                        pageStart = pageStartList[pageStartList.Count - 1];
                        pageLayout = htmlFile.GetPageLayout(pageStart, out nextPageStart, pageParams);

                        if (nextPageStart != null)
                           pageStartList.Add(nextPageStart);
                    }
                    while (!pageLayout.IsLastPage);

                    pageRequest.Location = new BookLocation(location.FileIndex, location.BreakIndex, pageStartList.Count - 1, pageLayout.CharacterOffset);
                    break;

                default:
                    throw new ArgumentException("PageProvider.GetPage: Unhandled PageRequestType of " + pageRequest.Type);
            }
            return pageLayout;
        }
    }
}
