﻿using System;
using System.IO;
using Windows.Foundation;
using Windows.Graphics.Display;
using Windows.Storage;
using Windows.UI.ViewManagement;
using Windows.UI.Xaml;
using Windows.UI.Xaml.Controls;
using Petzold.Epub;
using Petzold.Epub.Viewer.Windows8;

namespace TheNewEpublic
{
    public sealed partial class BookViewerPage
    {
        BookReference bookReference;
        bool wasLoaded;
        MemoryStream memoryStream;

        public BookViewerPage(BookReference bookReference)
        {
            InitializeComponent();
            this.bookReference = bookReference;
            Loaded += OnLoaded;
        }

        async void OnLoaded(object sender, RoutedEventArgs args)
        {
            if (wasLoaded)
                return;

            wasLoaded = true;

            // Transfer entire book file to MemoryStream
 	        StorageFile storageFile = await StorageFile.GetFileFromPathAsync(bookReference.FileName);
            byte[] buffer = await Helpers.ReadStorageFileAsync(storageFile);
            memoryStream = new MemoryStream(buffer, false);

            // Open as a ZIP file 
            ZipPackageAccessor packageAccessor = new ZipPackageAccessor(memoryStream);
            FontMetricsProvider fontMetricsProvider = new FontMetricsProvider();

            // Create EpubBook object
            EpubBook epubBook = new EpubBook(bookReference.OpfFileName, packageAccessor, fontMetricsProvider);
            PageTitle.Text = epubBook.BookInformation.Titles[0];

            // The Table of Contents is currently Collapsed
            contentsListBox.ItemsSource = epubBook.TableOfContents.Chapters;

            // Set up BookViewer and go to first page
            bookViewer.PageProvider = epubBook.GetPageProvider();
            bookViewer.PageTransition = new SlideTransition();              //  new SkewTransition();

            bookViewer.JumpToPage(bookReference.Location.FileIndex, 
                                  bookReference.Location.BreakIndex, 
                                  bookReference.Location.CharacterOffset);
            return;
        }

        async void OnBackButtonClick(object sender, RoutedEventArgs args)
        {
            // Finished with book -- dispose of MemoryStream, save location in book in AppSettings, and go back to MainPage
            memoryStream.Dispose();
            await(Application.Current as App).AppSettings.SaveAsync();
            Window.Current.Content = new MainPage();
        }

        // TODO: How about saving the AppSettings here?  Is that too much after each page flip?
        void OnBookViewerCurrentLocationChanged(object sender, object e)
        {
            bookReference.Location = bookViewer.CurrentLocation;
        }




        #region Automatic Stuff for Blank Page

        // View state management for switching among Full, Fill, Snapped, and Portrait states

        private DisplayPropertiesEventHandler _displayHandler;
        private TypedEventHandler<ApplicationLayout, ApplicationLayoutChangedEventArgs> _layoutHandler;

        private void Page_Loaded(object sender, RoutedEventArgs e)
        {
            if (_displayHandler == null)
            {
                _displayHandler = Page_OrientationChanged;
                _layoutHandler = Page_LayoutChanged;
            }
            DisplayProperties.OrientationChanged += _displayHandler;
            ApplicationLayout.GetForCurrentView().LayoutChanged += _layoutHandler;
            SetCurrentOrientation(this);
        }

        private void Page_Unloaded(object sender, RoutedEventArgs e)
        {
            DisplayProperties.OrientationChanged -= _displayHandler;
            ApplicationLayout.GetForCurrentView().LayoutChanged -= _layoutHandler;
        }

        private void Page_LayoutChanged(object sender, ApplicationLayoutChangedEventArgs e)
        {
            SetCurrentOrientation(this);
        }

        private void Page_OrientationChanged(object sender)
        {
            SetCurrentOrientation(this);
        }

        private void SetCurrentOrientation(Control viewStateAwareControl)
        {
            VisualStateManager.GoToState(viewStateAwareControl, this.GetViewState(), false);
        }

        private String GetViewState()
        {
            var orientation = DisplayProperties.CurrentOrientation;
            if (orientation == DisplayOrientations.Portrait ||
                orientation == DisplayOrientations.PortraitFlipped) return "Portrait";
            var layout = ApplicationLayout.Value;
            if (layout == ApplicationLayoutState.Filled) return "Fill";
            if (layout == ApplicationLayoutState.Snapped) return "Snapped";
            return "Full";
        }

        #endregion
    }
}
